/*
 * Autogenerated by the Meson build system.
 * Do not edit, your changes will be lost.
 */

#pragma once

#define RTE_ARCH i686

#define RTE_ARCH_32

#undef RTE_ARCH_64

#define RTE_ARCH_I686 1

#define RTE_ARCH_X86 1

#define RTE_BACKTRACE

#define RTE_BASEBAND_ACC 1

#define RTE_BASEBAND_FPGA_5GNR_FEC 1

#define RTE_BASEBAND_FPGA_LTE_FEC 1

#define RTE_BASEBAND_LA12XX 1

#define RTE_BASEBAND_NULL 1

#define RTE_BASEBAND_TURBO_SW 1

#define RTE_BUS_AUXILIARY 1

#define RTE_BUS_CDX 1

#define RTE_BUS_DPAA 1

#define RTE_BUS_FSLMC 1

#define RTE_BUS_IFPGA 1

#define RTE_BUS_PCI 1

#define RTE_BUS_PLATFORM 1

#define RTE_BUS_UACCE 1

#define RTE_BUS_VDEV 1

#define RTE_BUS_VMBUS 1

#define RTE_CACHE_LINE_SIZE 64

#define RTE_COMMON_CPT 1

#define RTE_COMMON_DPAAX 1

#define RTE_COMMON_IAVF 1

#define RTE_COMMON_IDPF 1

#define RTE_COMMON_IONIC 1

#define RTE_COMMON_NITROX 1

#define RTE_COMMON_QAT 1

#define RTE_COMPILE_TIME_CPUFLAGS RTE_CPUFLAG_SSE,RTE_CPUFLAG_SSE2,RTE_CPUFLAG_SSE3,RTE_CPUFLAG_SSSE3,RTE_CPUFLAG_SSE4_1,RTE_CPUFLAG_SSE4_2

#define RTE_COMPRESS_NITROX 1

#define RTE_COMPRESS_ZLIB 1

#define RTE_CRYPTO_BCMFS 1

#define RTE_CRYPTO_CAAM_JR 1

#define RTE_CRYPTO_CCP 1

#define RTE_CRYPTO_DPAA2_SEC 1

#define RTE_CRYPTO_DPAA_SEC 1

#define RTE_CRYPTO_IONIC 1

#define RTE_CRYPTO_NITROX 1

#define RTE_CRYPTO_NULL 1

#define RTE_CRYPTO_OPENSSL 1

#define RTE_CRYPTO_SCHEDULER 1

#define RTE_CRYPTO_VIRTIO 1

#define RTE_DMA_DPAA 1

#define RTE_DMA_DPAA2 1

#define RTE_DMA_IDXD 1

#define RTE_DMA_IOAT 1

#define RTE_DMA_SKELETON 1

#define RTE_DRIVER_MEMPOOL_BUCKET_SIZE_KB 64

#define RTE_EAL_NUMA_AWARE_HUGEPAGES 1

#define RTE_EAL_PMD_PATH "/usr/lib/dpdk/pmds-24.2"

#define RTE_EAL_VFIO

#undef RTE_ENABLE_STDATOMIC

#undef RTE_ENABLE_TRACE_FP

#define RTE_ENV_FREEBSD 0

#define RTE_ENV_LINUX 1

#define RTE_ENV_WINDOWS 2

#define RTE_EVENT_DPAA 1

#define RTE_EVENT_DPAA2 1

#define RTE_EVENT_DSW 1

#define RTE_EVENT_OPDL 1

#define RTE_EVENT_SKELETON 1

#define RTE_EVENT_SW 1

#define RTE_EXEC_ENV 1

#define RTE_EXEC_ENV_IS_FREEBSD 0

#define RTE_EXEC_ENV_IS_LINUX 1

#define RTE_EXEC_ENV_IS_WINDOWS 0

#define RTE_EXEC_ENV_LINUX 1

#define RTE_HAS_JANSSON 1

#define RTE_HAS_LIBARCHIVE 1

#define RTE_HAS_LIBFDT 1

#define RTE_HAS_LIBNUMA 1

#define RTE_HAS_LIBPCAP 1

#define RTE_HAS_OPENSSL 1

#define RTE_IOVA_IN_MBUF 1

#undef RTE_LIBEAL_USE_HPET

#define RTE_LIBRTE_DPAA2_USE_PHYS_IOVA

#define RTE_LIBRTE_VHOST_NUMA 1

#define RTE_LIBRTE_VHOST_POSTCOPY

#define RTE_LIB_ACL 1

#define RTE_LIB_ARGPARSE 1

#define RTE_LIB_BBDEV 1

#define RTE_LIB_BITRATESTATS 1

#define RTE_LIB_CFGFILE 1

#define RTE_LIB_CMDLINE 1

#define RTE_LIB_COMPRESSDEV 1

#define RTE_LIB_CRYPTODEV 1

#define RTE_LIB_DISPATCHER 1

#define RTE_LIB_DISTRIBUTOR 1

#define RTE_LIB_DMADEV 1

#define RTE_LIB_EAL 1

#define RTE_LIB_EFD 1

#define RTE_LIB_ETHDEV 1

#define RTE_LIB_EVENTDEV 1

#define RTE_LIB_FIB 1

#define RTE_LIB_GPUDEV 1

#define RTE_LIB_GRAPH 1

#define RTE_LIB_GRO 1

#define RTE_LIB_GSO 1

#define RTE_LIB_HASH 1

#define RTE_LIB_IPSEC 1

#define RTE_LIB_IP_FRAG 1

#define RTE_LIB_JOBSTATS 1

#define RTE_LIB_KVARGS 1

#define RTE_LIB_LATENCYSTATS 1

#define RTE_LIB_LOG 1

#define RTE_LIB_LPM 1

#define RTE_LIB_MBUF 1

#define RTE_LIB_MEMBER 1

#define RTE_LIB_MEMPOOL 1

#define RTE_LIB_METER 1

#define RTE_LIB_METRICS 1

#define RTE_LIB_MLDEV 1

#define RTE_LIB_NET 1

#define RTE_LIB_NODE 1

#define RTE_LIB_PCAPNG 1

#define RTE_LIB_PCI 1

#define RTE_LIB_PDCP 1

#define RTE_LIB_PIPELINE 1

#define RTE_LIB_PORT 1

#define RTE_LIB_POWER 1

#define RTE_LIB_PTR_COMPRESS 1

#define RTE_LIB_RAWDEV 1

#define RTE_LIB_RCU 1

#define RTE_LIB_REGEXDEV 1

#define RTE_LIB_REORDER 1

#define RTE_LIB_RIB 1

#define RTE_LIB_RING 1

#define RTE_LIB_SCHED 1

#define RTE_LIB_SECURITY 1

#define RTE_LIB_STACK 1

#define RTE_LIB_TABLE 1

#define RTE_LIB_TELEMETRY 1

#define RTE_LIB_TIMER 1

#define RTE_LIB_VHOST 1

#define RTE_MACHINE corei7

#define RTE_MAX_ETHPORTS 32

#define RTE_MAX_LCORE 128

#define RTE_MAX_MEM_MB 2048

#define RTE_MAX_NUMA_NODES 32

#define RTE_MAX_VFIO_GROUPS 64

#define RTE_MBUF_REFCNT_ATOMIC

#define RTE_MEMPOOL_BUCKET 1

#define RTE_MEMPOOL_DPAA 1

#define RTE_MEMPOOL_DPAA2 1

#define RTE_MEMPOOL_RING 1

#define RTE_MEMPOOL_STACK 1

#define RTE_NET_AF_PACKET 1

#define RTE_NET_ARK 1

#define RTE_NET_ATLANTIC 1

#define RTE_NET_AVP 1

#define RTE_NET_AXGBE 1

#define RTE_NET_BNX2X 1

#define RTE_NET_BNXT 1

#define RTE_NET_BOND 1

#define RTE_NET_CPFL 1

#define RTE_NET_CXGBE 1

#define RTE_NET_DPAA 1

#define RTE_NET_DPAA2 1

#define RTE_NET_E1000 1

#define RTE_NET_ENA 1

#define RTE_NET_ENETC 1

#define RTE_NET_ENETFEC 1

#define RTE_NET_ENIC 1

#define RTE_NET_FAILSAFE 1

#define RTE_NET_FM10K 1

#define RTE_NET_GVE 1

#define RTE_NET_HINIC 1

#define RTE_NET_I40E 1

#define RTE_NET_IAVF 1

#define RTE_NET_ICE 1

#define RTE_NET_IDPF 1

#define RTE_NET_IGC 1

#define RTE_NET_IONIC 1

#define RTE_NET_IPN3KE 1

#define RTE_NET_IXGBE 1

#define RTE_NET_MEMIF 1

#define RTE_NET_NETVSC 1

#define RTE_NET_NGBE 1

#define RTE_NET_NULL 1

#define RTE_NET_OCTEON_EP 1

#define RTE_NET_PCAP 1

#define RTE_NET_PFE 1

#define RTE_NET_QEDE 1

#define RTE_NET_RING 1

#define RTE_NET_SOFTNIC 1

#define RTE_NET_TAP 1

#define RTE_NET_TXGBE 1

#define RTE_NET_VDEV_NETVSC 1

#define RTE_NET_VHOST 1

#define RTE_NET_VIRTIO 1

#define RTE_NET_VMXNET3 1

#define RTE_PKTMBUF_HEADROOM 128

#define RTE_PORT_PCAP 1

#define RTE_QAT_OPENSSL

#define RTE_RAW_DPAA2_CMDIF 1

#define RTE_RAW_IFPGA 1

#define RTE_RAW_NTB 1

#define RTE_RAW_SKELETON 1

#define RTE_USE_LIBBSD 1

#define RTE_VDPA_IFC 1

#define RTE_VER_MINOR 0

#define RTE_VER_MONTH 7

#define RTE_VER_RELEASE 99

#define RTE_VER_SUFFIX ""

#define RTE_VER_YEAR 24

